/* Try to avoid some annoying page-breaking */
.no_break_inside, h1, h2, h3, li, p {
    page-break-inside: avoid !important;
}

h1, h2, h3 {
    margin:0;
    padding:0;
    page-break-after: avoid !important;
}

.page_break {
    page-break-after: always;
}

body {
    text-align:justify;
    padding: 0;
    margin: 0;
    font-family: "Computer Modern";
    font-size: 11pt;
}

.title, .author, .header, .footer {
    font-family: "Computer Modern";
}

.korean {
    font-family: "UnBatang";
}

.japanese {
    font-family: "IPA X0208 Pゴシック","IPAX0208PGothic";
}

.chinese, .chinese-traditional {
    font-family: "WenQuanYi Micro Hei";
}

p, li {
    line-height: 14.4pt;
}

.title {
    width:100%;
    text-align:center;
    font-size: 20.74pt;
    padding-top:25mm;
}

.author {
    width:100%;
    text-align:center;
    font-size: 14.4pt;
}

.spacer {
    display:block;
    height: 1cm;
}

a {
    color:black;
}

h2 a {
    text-decoration:none;
}

h1 {
    font-size: 17.28pt;
}

h2 {
    font-size: 14.4pt;
    padding-top: 0.5em;
    padding-bottom: 0.5em;
}

h3 {
    font-size: 12pt;
}

li {
    list-style: disc;
    padding: 0;
    margin-left: 2.2em;
}

li li {
    list-style: none;
    padding: 0;
    margin:0;
}

/* Custom li style derivated from here: */
/* http://stackoverflow.com/a/13354689 */
ul {
    padding: 0;
    margin-left: 1em;
    margin-top: .45em;
    margin-bottom: .45em;
}

ul li {
    list-style: none;
    padding-left: 1em;
}

ul ul li {
    padding-left: 2em;
}

ul li:before {
    /* This is the long dash */
    content: "\2013";
    display: inline-block;
    margin-left: -1em; /* same as padding-left set on li */
    width: 1em; /* same as padding-left set on li */
}

/*Breaking the hierarchy to improve page-breaking forces us to duplicate the style*/
ul ul li:before {
    /* This is '*' */
    content: "\2217";
    display: inline-block;
    margin-left: -1em; /* same as padding-left set on li */
    width: 1em; /* same as padding-left set on li */
}

